var ContentManagementComponentType = {
   CURRENT: "CURRENT",
   DELETED: "DELETED",
   HISTORY: "HISTORY"
};

var ContentManagementComponent = {}

ContentManagementComponent.MODEL_BY_ENTRY_ID = {};

ContentManagementComponent.savedMessages = null;

ContentManagementComponent.getForm = function()
{
   return document.forms[0];
}

ContentManagementComponent.refresh = function(instanceId, notifyChanged, clearOtherMsgs)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var form = ContentManagementComponent.getForm();
   var parameters = ContentManagementComponent.getTableParameters(instanceId, form);
   
   CommonService.refreshContentTable(
         model.serializableModel, parameters,
         ContentManagementComponent.createPageModificationHandler(model, notifyChanged, clearOtherMsgs)
      );
   
   return false;
}

ContentManagementComponent.getTableParameters = function(instanceId, form)
{
   if (!form)
   {
      form = ContentManagementComponent.getForm();
   }
   
   var params = getParamsForTable(form, instanceId);
   
   /*
    * Add any required parameters that were not in the model and table
    * parameters... For now, this is only the fileName parameter
    * (which should probably be in the model...)
    */
   var input = form.elements["fileName"];
   if (input)
   {
      params["fileName"] = input.value;
   }
   
   return params;
}

ContentManagementComponent.createPageModificationHandler = function(model, notifyChanged, clearOtherMsgs)
{
   var clearOtherMsgs = (clearOtherMsgs === undefined) ? (notifyChanged === true)
                                                       : (clearOtherMsgs === true);
   
   if (!notifyChanged && !clearOtherMsgs)
   {
      return ContentManagementComponent.handlePageModification;
   }
   
   var _model = model;
   
   var handleChange = function(pageModification)
   {
      ContentManagementComponent.handlePageModification(
            pageModification, clearOtherMsgs /* clearOtherMsgs */);
      
      /* Notify CPB backend that this field changed... */
      var orig_window = ContentManagementComponent.getOriginalContentManagementComponentWindow();
      if (notifyChanged && _model.fieldName && orig_window.cpb)
      {
         orig_window.cpb.fieldValueChanged(_model.fieldName);
      }
      
      /* refresh the original content management component (if different) */
      if (orig_window.ContentManagementComponent !== ContentManagementComponent)
      {
         orig_window.ContentManagementComponent.refresh(_model.instanceId,
               false /* notifyChanged */,
               true /* clearOtherMsgs */);
      }
   }
   
   return handleChange;
}

ContentManagementComponent.handlePageModification = function(pageModification, clearOtherMsgs)
{
   if (pageModification)
   {
      var orig_clearOtherMsgs = cpb.getClearOtherMessages();
      
      cpb.setClearOtherMessages(true);
      PageModification.handle(pageModification);
      cpb.setClearOtherMessages(orig_clearOtherMsgs);
      
      if (ContentManagementComponent.savedMessages)
      {
         /*
          * We get into this block if messages were "saved" in
          * ContentDialogClose.jsp. Since we forced the clearing
          * of other messages when handling the PageModification
          * we need to add the saved messages back (but don't clear
          * the ones that were just added by handling the PageModification).
          */
         cpb.displayMessages(cpb.INFO_MESSAGES_TYPE,
               ContentManagementComponent.savedMessages, false /* clearOtherMsgs */);
         ContentManagementComponent.savedMessages = null;
      }
   }
}

ContentManagementComponent.putContent = function(model, content)
{
   if (!model.contentByIndexMap)
   {
      model.contentByIndexMap = {};
   }
   
   model.contentByIndexMap[content.contentIndex] = content;
}

ContentManagementComponent.getContent = function(model, contentIndex)
{
   if (!model.contentByIndexMap)
   {
      return;
   }
   
   return model.contentByIndexMap[contentIndex];
}

ContentManagementComponent.getModelByInstanceId = function(instanceId)
{
   return ContentManagementComponent.MODEL_BY_ENTRY_ID[instanceId];
}

ContentManagementComponent.getQueryString = function(model)
{
   var queryString = new QueryString();
   
   for (var key in model.serializableModel)
   {
      var value = model.serializableModel[key];
      queryString.setParameter(key, value);
   }
   
   return queryString;
}

ContentManagementComponent.addContent = function(instanceId)
{
   ContentManagementComponent.showContent(instanceId);
}

ContentManagementComponent.showContent = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   
   var queryString = ContentManagementComponent.getQueryString(model);
   
   var content = null;
   var fileName = null;
   var title;
   var height = 600;
   
   if (contentIndex !== undefined)
   {
      content = ContentManagementComponent.getContent(model, contentIndex);
      queryString.setParameter("fileName", content.fileName);
      
      if (content.versionName)
      {
         queryString.setParameter("versionName", content.versionName);
         title = AppProperties["contentManagement.viewContent.title.txt"];
      }
      else
      {
         if (model.readOnly)
         {
            title = AppProperties["contentManagement.viewContent.title.txt"];
         }
         else
         {
            if (model.type == ContentManagementComponentType.DELETED)
            {
               title = AppProperties["contentManagement.viewContent.title.txt"];
            }
            else
            {
               title = AppProperties["contentManagement.editContent.title.txt"];
            }
         }
      }
   }
   else
   {
      title = AppProperties["contentManagement.addContent.title.txt"];
      height = 450;
   }
   
   IFrameDialog.show({
      title: title,
      url: "ContentDialog?method=showContent&" + queryString.toString(),
      width: 700,
      height: height
      });
}

ContentManagementComponent.defaultAction = function(handle)
{
   handle = decodeURI(handle);
   var pos = handle.indexOf("|");
   var instanceId = handle.substring(0, pos);
   var fileName;
   var version;
   
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   if (handle.lastIndexOf("|") > pos)
   {
      // HANDLE FORMAT:
      // instanceId|fileName|versionName
      lastPos = handle.lastIndexOf("|");
      
      fileName = handle.substring(pos+1, lastPos);
      versionName = handle.substring(lastPos+1);
   }
   else
   {
      // HANDLE FORMAT:
      // instanceId|fileName
      fileName = handle.substring(pos+1);
      versionName = null;
   }
   
   ContentManagementComponent.downloadContent(model, {
      fileName: fileName,
      versionName: versionName
      });
}

ContentManagementComponent.downloadContent = function(instanceIdOrModel, contentOrIndex)
{
   var model;
   var content;
   
   if (instanceIdOrModel.constructor === String)
   {
      // instanceIdOrModel is instanceId
      model = ContentManagementComponent.getModelByInstanceId(instanceIdOrModel);
      // contentOrIndex is contentIndex
      content = ContentManagementComponent.getContent(model, contentOrIndex);
   }
   else
   {
      // instanceIdOrModel is model
      model = instanceIdOrModel;
      // contentOrIndex is content
      content = contentOrIndex;
   }
   
   var url = new URL("Content");
   var queryString = url.getQueryString();
   
   queryString.setParameter("method", "downloadContent");
   queryString.setParameter("businessObjectRef", model.businessObjectRef);
   queryString.setParameter("fileName", content.fileName);
   queryString.setParameter("versionName", content.versionName);
   
   window.location = url.toString();
}

ContentManagementComponent.unlockContent = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var content = ContentManagementComponent.getContent(model, contentIndex);
   var fileName = content.fileName;
   
   CommonService.unlockContent(
         model.serializableModel, fileName,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.lockContent = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var content = ContentManagementComponent.getContent(model, contentIndex);
   var fileName = content.fileName;
   
   CommonService.lockContent(
         model.serializableModel, fileName,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.deleteContent = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var content = ContentManagementComponent.getContent(model, contentIndex);
   var fileName = content.fileName;
   
   if (!confirm(AppProperties["contentManagement.delete.confirm.txt"]))
   {
      return;
   }
   
   CommonService.deleteContent(
         model.serializableModel, fileName,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.undeleteContent = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var content = ContentManagementComponent.getContent(model, contentIndex);
   var fileName = content.fileName;
   
   if (!confirm(AppProperties["contentManagement.undelete.confirm.txt"]))
   {
      return;
   }
   
   CommonService.undeleteContent(
         model.serializableModel, fileName,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.deleteSelectedContent = function(instanceId, fileNameList)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   
   var form = ContentManagementComponent.getForm();
   var checkboxList = form.elements[instanceId + ".fileNameList"];
   if (!checkboxList)
   {
      alert(AppProperties["contentManagement.noneSelected.txt"]);
      return;
   }
   
   if (checkboxList.nodeName)
   {
      checkboxList = [checkboxList];
   }
   
   var fileNameList = [];
   
   for (var i = 0; i < checkboxList.length; i++)
   {
      var checkbox = checkboxList[i];
      
      if (checkbox.checked)
      {
         fileNameList.push(checkbox.value);
      }
   }
   
   if (fileNameList.length == 0)
   {
      alert(AppProperties["contentManagement.noneSelected.txt"]);
      return;
   }
   
   if (!confirm(AppProperties["contentManagement.deleteSelected.confirm.txt"]))
   {
      return;
   }
   
   CommonService.deleteSelectedContent(
         model.serializableModel, fileNameList,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.undeleteSelectedContent = function(instanceId, fileNameList)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   
   var form = ContentManagementComponent.getForm();
   var checkboxList = form.elements[instanceId + ".fileNameList"];
   if (!checkboxList)
   {
      alert(AppProperties["contentManagement.noneSelected.txt"]);
      return;
   }
   
   if (checkboxList.nodeName)
   {
      checkboxList = [checkboxList];
   }
   
   var fileNameList = [];
   
   for (var i = 0; i < checkboxList.length; i++)
   {
      var checkbox = checkboxList[i];
      
      if (checkbox.checked)
      {
         fileNameList.push(checkbox.value);
      }
   }
   
   if (fileNameList.length == 0)
   {
      alert(AppProperties["contentManagement.noneSelected.txt"]);
      return;
   }
   
   if (!confirm(AppProperties["contentManagement.undeleteSelected.confirm.txt"]))
   {
      return;
   }
   
   CommonService.undeleteSelectedContent(
         model.serializableModel, fileNameList,
         ContentManagementComponent.getTableParameters(instanceId),
         ContentManagementComponent.createPageModificationHandler(model, true /* notifyChanged */)
      );
}

ContentManagementComponent.showContextMenu = function(content) {

   var instanceId = content.instanceId;
   var contentIndex = content.contentIndex;
   
   var model = ContentManagementComponent.getModelByInstanceId(content.instanceId);
   
   var readOnly = model.readOnly;
   var typeDeleted = (model.type == ContentManagementComponentType.DELETED);
   var typeHistorical = (model.type == ContentManagementComponentType.HISTORY);
   var typeCurrent = (model.type == ContentManagementComponentType.CURRENT);
   
   var menu = new TableMenu();
   
   ContentManagementComponent.putContent(model, content);
   
   menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
         ThemedImage.ATTACHMENTS.url, 
         AppProperties["application.download.txt"],
         "javascript: ContentManagementComponent.downloadContent('" + instanceId + "', " + contentIndex + ")", 
         true));
   
   if (!typeHistorical)
   {
      menu.addMenuItem(new MenuItem(MenuItem.prototype.separator));
      
      if (!readOnly)
      {
         if (content.canLock)
         {
            menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
                  ThemedImage.LOCK_ENABLED.url,
                  AppProperties["application.lock.txt"],
                  "javascript: ContentManagementComponent.lockContent('" + instanceId + "', " + contentIndex + ")",
                  true));
         }
         
         if (content.canUnlock)
         {
            menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
                  ThemedImage.UNLOCK_ENABLED.url, 
                  AppProperties["application.unlock.txt"],
                  "javascript: ContentManagementComponent.unlockContent('" + instanceId + "', " + contentIndex + ")",
                  true));
         }
      }
      
      menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
            Application.CONTEXT_PATH + '/' + AppProperties["application.history.image"], 
            AppProperties["historyButton.txt"],
            "javascript: ContentManagementComponent.viewContentHistory('" + instanceId + "', " + contentIndex + ")",
            true));
      
      if (typeDeleted)
      {
         menu.addMenuItem(new MenuItem(MenuItem.prototype.separator));
         
         menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
               ThemedImage.UNDELETE.url, 
               AppProperties["application.restore.txt"],
               "javascript: ContentManagementComponent.undeleteContent('" + instanceId + "', " + contentIndex + ")",
               true));
      }
      else if (!readOnly && typeCurrent && content.canDelete)
      {
         menu.addMenuItem(new MenuItem(MenuItem.prototype.separator));
         
         menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
               ThemedImage.DELETE.url, 
               AppProperties["application.delete.txt"],
               "javascript: ContentManagementComponent.deleteContent('" + instanceId + "', " + contentIndex + ")",
               true));
      }
   }
   
   menu.addMenuItem(new MenuItem(MenuItem.prototype.separator));
   
   menu.addMenuItem(new MenuItem(MenuItem.prototype.action,
         ThemedImage.EDIT.url, 
         AppProperties["application.properties.txt"],
         "javascript: ContentManagementComponent.showContent('" + instanceId + "', " + contentIndex + ")", 
         true));
   
   menu.render();
}

ContentManagementComponent.viewContentHistory = function(instanceId, contentIndex)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   var content = ContentManagementComponent.getContent(model, contentIndex);
   
   var queryString = ContentManagementComponent.getQueryString(model);
   queryString.setParameter("type", ContentManagementComponentType.HISTORY);
   queryString.setParameter("fileName", content.fileName);
   
   var title = AppProperties["contentManagement.contentHistory.title.txt"];
   
   ContentManagementComponent.historyAttachmentsDialog = IFrameDialog.show({
      title: title,
      url: "HistoryContent?method=showHistoryContent&" + queryString.toString(),
      width: 750,
      height: 550,
      name: "historyContent"
      });
}

ContentManagementComponent.viewAllHistory = function(instanceId)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   
   var queryString = ContentManagementComponent.getQueryString(model);
   queryString.setParameter("type", ContentManagementComponentType.HISTORY);
   
   var title = AppProperties["contentManagement.contentHistory.title.txt"];
   
   ContentManagementComponent.historyAttachmentsDialog = IFrameDialog.show({
      title: title,
      url: "HistoryContent?method=showHistoryContent&" + queryString.toString(),
      width: 750,
      height: 550,
      name: "historyContent"
      });
}

ContentManagementComponent.showAllDeletedContent = function(instanceId)
{
   var model = ContentManagementComponent.getModelByInstanceId(instanceId);
   
   var queryString = ContentManagementComponent.getQueryString(model);
   var title = AppProperties["contentManagement.deletedContent.title.txt"];
   
   if (model.readOnly)
   {
      queryString.setParameter("viewMode", ViewMode.READ_ONLY);
   }
   else
   {
      queryString.setParameter("viewMode", ViewMode.EDITABLE);
   }
   
   queryString.setParameter("type", ContentManagementComponentType.DELETED);
   
   ContentManagementComponent.deletedAttachmentsDialog = IFrameDialog.show({
      title: title,
      url: "DeletedContent?method=showDeletedContent&" + queryString.toString(),
      width: 750,
      height: 550,
      name: "deletedContent"
      });
}

ContentManagementComponent.getActiveContentManagementComponentWindow = function()
{
   var dialog = IFrameDialog.getActiveDialog();
   
   while(dialog) 
   {
      if (dialog.iframe)
      {
         var dialogWindow = dialog.iframe.contentWindow;
         if (dialogWindow.ContentManagementComponent)
         {
            return dialogWindow;
         }
      }
      
      dialog = dialog.parent;
   }
   
   return window;
}

ContentManagementComponent.getOriginalContentManagementComponentWindow = function()
{
   if (window.parent && window.parent.ContentManagementComponent)
   {
      return window.parent;
   }
   else
   {
      return window;
   }
}