/***************************************************************************/
/*  Copyright (c) 2005-2014 by SAS Institute, Inc., Cary, NC, USA.         */
/*  All rights reserved.                                                   */
/***************************************************************************/
/*  NAME:        orm_lasr_perm_filter.sas                                  */
/*  VERSION:     6.1                                                       */
/*  DESCRIPTION: Filters the LASR table by feature and user.               */
/*  PRODUCT:     SAS Enterprise GRC                                        */
/***************************************************************************/

%macro orm_lasr_perm_filter;

%put NOTE: {START: &sysmacroname};

%macro get_user_sec_isUnder_constraints(feature=,
                                        user_id=,
                                        security_acl_lib=);

%put NOTE: {START: &sysmacroname};

  %get_known_isUnder_constraints(feature=&feature,
                                 sectable=&security_acl_lib..sm_security_acl(where=(user_id="&user_id" and business_object_nm="&feature")))

%put NOTE: {END: &sysmacroname};

%mend get_user_sec_isUnder_constraints;


%macro get_known_isUnder_constraints(feature=,
                                     sectable=);

%put NOTE: {START: &sysmacroname};

  %local dsid rc oneline;
  %global acc;

  /* open the security table and iterate through it */
  %let dsid=%sysfunc(open(&sectable,i));
  %if &dsid eq 0 %then %do;
    /* bail */
  %end;
  %let acc=;
  %let rc=%sysfunc(fetch(&dsid));
  %do %while(&rc eq 0);
    %let oneline=%known_isUnder_constraints_point(feature=&feature, dsid=&dsid);
    /* add the constraints for the current point using or */
   %if %length(&acc) ne 0 %then %let acc=%cmpres(&acc) or;
   %if %nrbquote(&oneline) eq  %then %let acc=;
   %else %if %nrbquote(&oneline)=1 or %nrbquote(&oneline)=0 %then %let acc=%cmpres(&acc) %cmpres(&oneline);
   %else %let acc=%cmpres(&acc) %cmpres((&oneline));
    %let rc=%sysfunc(fetch(&dsid));
  %end;
  %let rc=%sysfunc(close(&dsid));

%put NOTE: {END: &sysmacroname};

%mend get_known_isUnder_constraints;


%macro known_isUnder_constraints_point(feature=,
                                       dsid=);

%put NOTE: {START: &sysmacroname};

  %local i j f dimension levels featureDims;
  %let featureDims=%dims_for_feature(feature=&feature);

  %let i = 1;
  %let dimension = %scan(&distinct_dims,&i,'*');
  %let levels = %scan(&distinct_dims_levels,&i,'*');
  %let line=;
  %do %while (&dimension ne);
      %let isunder = %known_isUnder_constraints_oneDim(feature=&feature, dsid=&dsid,
                       dimension=&dimension, levels=&levels, featureDims=&featureDims);
      /* avoid where clause with a bunch of 1 and 1 and 1 and 1... */
      %if %nrbquote(&isunder) ne 1 %then
      %do;
        %if %nrbquote(&line) ne %then %let line = &line %str()and%str();
        %let line = &line &isunder;
      %end;
    %let i = %eval(&i + 1);
    %let dimension = %scan(&distinct_dims,&i,'*');
    %let levels = %scan(&distinct_dims_levels,&i,'*');
  %end;

  %if %nrbquote(&line) ne  %then &line

%put NOTE: {END: &sysmacroname};

%mend known_isUnder_constraints_point;

%macro known_isUnder_constraints_oneDim(feature=,
                                        dsid=,
                                        dimension=,
                                        levels=,
                                        featureDims=);

%put NOTE: {START: &sysmacroname};

  %local fieldname fieldval i;
  %let fieldname=&dimension._RK;
  %let fieldval=%nrbquote(%sysfunc(getvarn(&dsid, %sysfunc(varnum(&dsid,&fieldname)))));
  %if &fieldval eq . %then 1 ;
  %else %do;
    %if %index(&featureDims,&dimension) eq 0 %then 0;
    %else %if &fieldval eq -1 %then %do;
         /* for the dimension root, match any node in the dimension, including the root */
        %nrbquote(&fieldname) ne "."
    %end;
    %else %do;
      %let i=1;
      %let fieldname=&dimension._RK_&i;
      %let fieldval=%nrbquote(%sysfunc(putn(%sysfunc(getvarn(&dsid, %sysfunc(varnum(&dsid,&fieldname)))),&dimension.fmt.)));
       %do %while(&i <= &levels and (%nrbquote(&fieldval) ne %str() and %nrbquote(&fieldval) ne %str(.)));
        &fieldname = "&fieldval"
        %let i=%eval(&i+1);
        %if (&i <= &levels) %then
        %do;
          %let fieldname=&dimension._RK_&i;
          %let fieldval=%nrbquote(%sysfunc(putn(%sysfunc(getvarn(&dsid, %sysfunc(varnum(&dsid,&fieldname)))),&dimension.fmt.)));
          %if (%nrbquote(&fieldval) ne %str() and %nrbquote(&fieldval) ne %str(.)) %then and ;
        %end;
      %end;
    %end;
  %end;

%put NOTE: {END: &sysmacroname};

%mend known_isUnder_constraints_oneDim;

%macro dims_for_feature(feature=);

%put NOTE: {START: &sysmacroname};

  %local result i f dimension;
  %let result=*;
  %let i = 1;
  %let f = %scan(&orm_features,&i,'*');
  %let dimension = %scan(&orm_dimensions,&i,'*');
  %do %while (&f ne);
    %if &f = &feature %then %do;
      %let result=&result.&dimension.*;
    %end;
    %let i = %eval(&i + 1);
    %let f = %scan(&orm_features,&i,'*');
    %let dimension = %scan(&orm_dimensions,&i,'*');
  %end;
  &result

%put NOTE: {END: &sysmacroname};

%mend dims_for_feature;

%put NOTE: {END: &sysmacroname};

%mend orm_lasr_perm_filter;
