/*******************************************************************************/
/* Copyright (c) 2005-2014 by SAS Institute Inc., Cary, NC, USA.               */
/* All Rights Reserved.                                                        */
/*******************************************************************************/
/* NAME:          orm_pivot_datatype                                           */
/* VERSION:       6.1                                                          */
/* DESCRIPTION:   add custom fields to a business object in the opdetail mart  */
/*                                                                             */
/*                                                                             */
/* PRODUCT:       SAS Enterprise GRC                                           */
/* USAGE:                                                                      */
/* PARAMETERS:    dtfmt: SAS date format and informat (default NLDATE.)        */
/*                datatype: type of data to create (business_object_type_nm)   */
/*                in_tablenm: input table name (typically, business_object_nm) */
/*                out_tablenm: output table name (optional, used when you want */
/*                             to use a different name for the output table nm)*/
/*                debug: N for removing all temp tables and files (default N)  */
/*******************************************************************************/
%macro orm_generateHistoryTable(rk_column=,
                                in_table=,
                                bo_name=,
                                outputtable=);
                            
%put NOTE: {START: &sysmacroname};

   %let QUOTESTR=%str(%');
   %let temptablenm=t_&in_table;
   %let doubletemptablenm=t_&temptablenm;
   %let historytablenm=&outputtable;
    %if %eval(%sysfunc(exist(sasoprsk.&temptablenm)) = 1) %then 
     %do;
      proc sql;
        drop table sasoprsk.&temptablenm;
      quit;
     %end;
    %if %eval(%sysfunc(exist(sasoprsk.&doubletemptablenm)) = 1) %then 
     %do;
      proc sql;
        drop table sasoprsk.&doubletemptablenm;
      quit;
     %end;
    %if %eval(%sysfunc(exist(sasoprsk.&historytablenm)) = 1) %then 
     %do;
      proc sql;
        drop table sasoprsk.&historytablenm;
      quit;
     %end;
       
   /*We have to use SAS's capability to execute native sql here because of syntax limitations in their validation.*/
   proc sql;
   %orm_db_connect;
   &DBConnectSQL;
   execute(
   %if %upcase(&orm_dbserver) = SQLSERVER %then
   %do;
      select mmm.* into &temptablenm from (
   %end;
   %else
   %do;
      CREATE
       TABLE &temptablenm as
       %if %upcase(&orm_dbserver) = DB2 %then
       %do;
         (
        %end;
   %end;
        SELECT
                G.*
                , ROW_NUMBER() OVER (
                    PARTITION BY
                        &rk_column
                        ,record_version_no
                    ORDER BY
                        &rk_column
                        ,record_version_no
                        ,transition_start_dttm
                ) AS group_num
            FROM
                (
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,valid_from_dttm AS transition_start_dttm
                        FROM
                            &in_table
                    UNION
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,transition_start_dttm
                        FROM
                            (
                                SELECT
                                        A.&rk_column
                                        ,A.valid_from_dttm
                                        ,A.valid_to_dttm
                                        ,A.record_version_no
                                        ,B.cust_field_nm
                                        ,B.value_no
                                        ,B.value_dt
                                        ,B.valid_from_dttm AS transition_start_dttm
                                        ,'UPDATE' AS action
                                        ,B.cust_field_val_rk
                                        ,B.record_version_no AS cust_field_record_version_no
                                    FROM
                                        &in_table A
                                        ,cust_field_value B
                                    WHERE
                                        (
                                            B.valid_from_dttm >= A.valid_from_dttm
                                            AND B.valid_from_dttm < A.valid_to_dttm
                                        )
                                        AND A.&rk_column = B.business_object_rk
                                        AND b.business_object_nm = (select distinct business_object_nm from business_object_type where business_object_type_nm = &bo_name)
                              AND ((B.value_txt is not null 
                                             or B.value_no is not null
                                  or B.value_flg is not null
                                  or B.value_cd is not null
                                  or B.value_dt is not null) or  
                                 (B.value_txt is null 
                                             and B.value_no is null
                                  and B.value_flg is null
                                  and B.value_cd is null
                                  and B.value_dt is null
                                  and B.record_version_no != 1))
                                UNION ALL
                                SELECT
                                        A.&rk_column
                                        ,A.valid_from_dttm
                                        ,A.valid_to_dttm
                                        ,A.record_version_no
                                        ,B.cust_field_nm
                                        ,B.value_no
                                        ,B.value_dt
                                        ,B.valid_to_dttm AS transition_start_dttm
                                        ,'DELETE' AS action
                                        ,B.cust_field_val_rk
                                        ,B.record_version_no AS cust_field_record_version_no
                                    FROM
                                        &in_table A
                                        ,cust_field_value B
                                    WHERE
                                        (
                                            B.valid_to_dttm >= A.valid_from_dttm
                                            AND B.valid_to_dttm < A.valid_to_dttm
                                        )
                                        AND A.&rk_column = B.business_object_rk
                                        AND b.business_object_nm = (select distinct business_object_nm from business_object_type where business_object_type_nm = &bo_name)
                                        AND B.valid_to_dttm <> &ORM_INFINITY
                                        AND cust_field_val_rk NOT IN (
                                            SELECT
                                                    cust_field_val_rk
                                                FROM
                                                    cust_field_value_l
                                        )
                                        AND b.record_version_no = (
                                            SELECT
                                                    MAX(record_version_no)
                                                FROM
                                                    cust_field_value
                                                WHERE
                                                    cust_field_value.cust_field_val_rk = b.cust_field_val_rk
                                        ) 
                              AND (B.value_txt is not null 
                                             or B.value_no is not null
                                  or B.value_flg is not null
                                  or B.value_cd is not null
                                  or B.value_dt is not null)
                            ) X
                    UNION
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,valid_to_dttm AS transition_start_dttm
                        FROM
                            &in_table
                ) G GROUP BY
                            &rk_column
                            ,record_version_no
                            ,transition_start_dttm
                            
       %if %upcase(&orm_dbserver) = SQLSERVER %then
       %do;
             ) mmm
       %end;
       %else %if %upcase(&orm_dbserver) = DB2 %then
       %do;
             ) WITH NO DATA
       %end;

      ) by SASOPRSK_CONN;
      /* DB2 creates the temp table without data so we need to populate the table still. */
      %if %upcase(&orm_dbserver) = DB2 %then
      %do;
     execute (
      insert into &temptablenm SELECT
                G.*
                , ROW_NUMBER() OVER (
                    PARTITION BY
                        &rk_column
                        ,record_version_no
                    ORDER BY
                        &rk_column
                        ,record_version_no
                        ,transition_start_dttm
                ) AS group_num
            FROM
                (
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,valid_from_dttm AS transition_start_dttm
                        FROM
                            &in_table
                    UNION
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,transition_start_dttm
                        FROM
                            (
                                SELECT
                                        A.&rk_column
                                        ,A.valid_from_dttm
                                        ,A.valid_to_dttm
                                        ,A.record_version_no
                                        ,B.cust_field_nm
                                        ,B.value_no
                                        ,B.value_dt
                                        ,B.valid_from_dttm AS transition_start_dttm
                                        ,'UPDATE' AS action
                                        ,B.cust_field_val_rk
                                        ,B.record_version_no AS cust_field_record_version_no
                                    FROM
                                        &in_table A
                                        ,cust_field_value B
                                    WHERE
                                        (
                                            B.valid_from_dttm >= A.valid_from_dttm
                                            AND B.valid_from_dttm < A.valid_to_dttm
                                        )
                                        AND A.&rk_column = B.business_object_rk
                                        AND b.business_object_nm = (select distinct business_object_nm from business_object_type where business_object_type_nm = &bo_name)
                              AND ((B.value_txt is not null 
                                             or B.value_no is not null
                                  or B.value_flg is not null
                                  or B.value_cd is not null
                                  or B.value_dt is not null) or  
                                 (B.value_txt is null 
                                             and B.value_no is null
                                  and B.value_flg is null
                                  and B.value_cd is null
                                  and B.value_dt is null
                                  and B.record_version_no != 1))
                                UNION ALL
                                SELECT
                                        A.&rk_column
                                        ,A.valid_from_dttm
                                        ,A.valid_to_dttm
                                        ,A.record_version_no
                                        ,B.cust_field_nm
                                        ,B.value_no
                                        ,B.value_dt
                                        ,B.valid_to_dttm AS transition_start_dttm
                                        ,'DELETE' AS action
                                        ,B.cust_field_val_rk
                                        ,B.record_version_no AS cust_field_record_version_no
                                    FROM
                                        &in_table A
                                        ,cust_field_value B
                                    WHERE
                                        (
                                            B.valid_to_dttm >= A.valid_from_dttm
                                            AND B.valid_to_dttm < A.valid_to_dttm
                                        )
                                        AND A.&rk_column = B.business_object_rk
                                        AND b.business_object_nm = (select distinct business_object_nm from business_object_type where business_object_type_nm = &bo_name)
                                        AND B.valid_to_dttm <> &ORM_INFINITY
                                        AND cust_field_val_rk NOT IN (
                                            SELECT
                                                    cust_field_val_rk
                                                FROM
                                                    cust_field_value_l
                                        )
                                        AND b.record_version_no = (
                                            SELECT
                                                    MAX(record_version_no)
                                                FROM
                                                    cust_field_value
                                                WHERE
                                                    cust_field_value.cust_field_val_rk = b.cust_field_val_rk
                                        ) 
                              AND (B.value_txt is not null 
                                             or B.value_no is not null
                                  or B.value_flg is not null
                                  or B.value_cd is not null
                                  or B.value_dt is not null)
                            ) X
                    UNION
                    SELECT
                            &rk_column
                            ,record_version_no
                            ,valid_to_dttm AS transition_start_dttm
                        FROM
                            &in_table
                ) G GROUP BY
                            &rk_column
                            ,record_version_no
                            ,transition_start_dttm
     ) by SASOPRSK_CONN;
     %end;

      execute(
      %if %upcase(&orm_dbserver) = SQLSERVER %then
      %do;
        select mmm.* into &doubletemptablenm from (
      %end;
      %else
      %do;
        Create table &doubletemptablenm as
        %if %upcase(&orm_dbserver) = DB2 %then
        %do;
             (
        %end;
      %end;
         SELECT
            A.*
            ,B.transition_start_dttm AS transition_end_dttm
         FROM
            &temptablenm B
            ,&temptablenm A
         WHERE
            A.&rk_column = B.&rk_column
            AND A.record_version_no = B.record_version_no
            AND A.GROUP_NUM + 1 = B.GROUP_NUM
         %if %upcase(&orm_dbserver) = SQLSERVER %then
         %do;
             ) mmm
         %end;
         %else %if %upcase(&orm_dbserver) = DB2 %then
         %do;
             ) WITH NO DATA
         %end;
      ) by SASOPRSK_CONN;

     /*DB2 creates the table without data so we need to insert the records. */
     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
        execute(
      insert into &doubletemptablenm SELECT
            A.*
            ,B.transition_start_dttm AS transition_end_dttm
         FROM
            &temptablenm B
            ,&temptablenm A
         WHERE
            A.&rk_column = B.&rk_column
            AND A.record_version_no = B.record_version_no
            AND A.GROUP_NUM + 1 = B.GROUP_NUM
      
      ) by SASOPRSK_CONN;
     %end;

      execute(
      %if %upcase(&orm_dbserver) = SQLSERVER %then
      %do;
         select mmm.* into &historytablenm from (
      %end;
      %else
      %do;
            create table &historytablenm as
         %if %upcase(&orm_dbserver) = DB2 %then
         %do;
               (
         %end;
      %end;
         SELECT
               L.*
               ,ROW_NUMBER() OVER (
                  PARTITION BY
                      l.&rk_column
                  order by 
                    l.&rk_column,
                 transition_start_dttm
               ) AS transition_no
               ,transition_start_dttm
               ,transition_end_dttm
                ,(CASE WHEN transition_end_dttm = L.valid_to_dttm THEN L.change_reason_rk
              else (select min(B.change_reason_rk) from change_reason B where L.&rk_column = B.business_object_rk 
                    and B.save_dttm = transition_end_dttm 
                    and b.business_object_nm = (select business_object_nm from business_object_type 
                                                where business_object_type_nm = &bo_name)
                    ) 
                  END) as real_reason_rk
         FROM
               &in_table L
               ,&doubletemptablenm T
         WHERE
               l.&rk_column = t.&rk_column
            and l.RECORD_VERSION_NO = t.RECORD_VERSION_NO
         %if %upcase(&orm_dbserver) = SQLSERVER %then
         %do;
             ) mmm
         %end;
         %else %if %upcase(&orm_dbserver) = DB2 %then
         %do;
             ) WITH NO DATA
         %end;
      ) by SASOPRSK_CONN;
     

     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
       /*DB2 creates the table without data so we need to insert the records. */
      execute(
         insert into &historytablenm SELECT
               L.*
               ,ROW_NUMBER() OVER (
                  PARTITION BY
                      l.&rk_column
                  order by 
                    l.&rk_column,
                 transition_start_dttm
               ) AS transition_no
               ,transition_start_dttm
               ,transition_end_dttm
                ,(CASE WHEN transition_end_dttm = L.valid_to_dttm THEN L.change_reason_rk
              else (select min(B.change_reason_rk) from change_reason B where L.&rk_column = B.business_object_rk 
                    and B.save_dttm = transition_end_dttm 
                    and b.business_object_nm = (select business_object_nm from business_object_type 
                                                where business_object_type_nm = &bo_name)
                    ) 
                  END) as real_reason_rk
         FROM
               &in_table L
               ,&doubletemptablenm T
         WHERE
               l.&rk_column = t.&rk_column
            and l.RECORD_VERSION_NO = t.RECORD_VERSION_NO
       ) by SASOPRSK_CONN; 
       
      /*Special handling of renaming colunmns for DB2 because they don't support alter table column rename statements. */
      execute(
            update &historytablenm set valid_from_dttm = transition_start_dttm
       ) by SASOPRSK_CONN;    

      execute(
            update &historytablenm set valid_to_dttm = transition_end_dttm
       ) by SASOPRSK_CONN;    

      execute(
            update &historytablenm set record_version_no = transition_no
       ) by SASOPRSK_CONN;    

      execute(
            update &historytablenm set change_reason_rk = real_reason_rk
       ) by SASOPRSK_CONN;    

      execute(
            alter table &historytablenm drop column real_reason_rk
       ) by SASOPRSK_CONN;

      execute(
            alter table &historytablenm drop column transition_no
       ) by SASOPRSK_CONN;

      execute(
            alter table &historytablenm drop column transition_end_dttm
       ) by SASOPRSK_CONN;

      /*Execute the reorg for DB2 after each update to avoid getting the table in an invalid state.*/
      execute( CALL SYSPROC.ADMIN_CMD ( &QUOTESTR. REORG TABLE &historytablenm. &QUOTESTR.) ) by SASOPRSK_CONN;

      execute(
            alter table &historytablenm drop column transition_start_dttm
       ) by SASOPRSK_CONN;

     %end;

     %if %upcase(&orm_dbserver) ne DB2 %then
     %do;
         /* Column renaming for non-DB2 databases. */
         execute(
            alter table &historytablenm drop column valid_from_dttm
         ) by SASOPRSK_CONN;

        execute(
            alter table &historytablenm drop column valid_to_dttm
         ) by SASOPRSK_CONN;

        execute(
            alter table &historytablenm drop column record_version_no
         ) by SASOPRSK_CONN;

         execute(
            alter table &historytablenm drop column change_reason_rk
         ) by SASOPRSK_CONN;

         execute(
            %if %upcase(&orm_dbserver) = SQLSERVER %then
            %do;
               sp_rename &QUOTESTR.&historytablenm..transition_start_dttm&QUOTESTR., 'valid_from_dttm','COLUMN'
            %end;
            %else
            %do;
               alter table &historytablenm rename column transition_start_dttm to valid_from_dttm
            %end;
         ) by SASOPRSK_CONN;

         execute(
            %if %upcase(&orm_dbserver) = SQLSERVER %then
            %do;
               sp_rename &QUOTESTR.&historytablenm..transition_end_dttm&QUOTESTR., 'valid_to_dttm','COLUMN'
            %end;
            %else
            %do;
               alter table &historytablenm rename column transition_end_dttm to valid_to_dttm
            %end;
         ) by SASOPRSK_CONN;
        
         execute(
            %if %upcase(&orm_dbserver) = SQLSERVER %then
            %do;
               sp_rename &QUOTESTR.&historytablenm..transition_no&QUOTESTR., 'record_version_no','COLUMN'
            %end;
            %else
            %do;
               alter table &historytablenm rename column transition_no to record_version_no
            %end;
         ) by SASOPRSK_CONN;
        
         execute(
            %if %upcase(&orm_dbserver) = SQLSERVER %then
            %do;
               sp_rename &QUOTESTR.&historytablenm..real_reason_rk&QUOTESTR., 'change_reason_rk','COLUMN'
            %end;
            %else
            %do;
               alter table &historytablenm rename column real_reason_rk to change_reason_rk
            %end;
         ) by SASOPRSK_CONN;
     %end;
     
     /* Drop the temp tables but keep the history table around for future use. We'll have to clean that up later.*/
      execute(
         drop table &temptablenm
      ) by SASOPRSK_CONN;
      execute(
         drop table &doubletemptablenm
      ) by SASOPRSK_CONN;
      disconnect from SASOPRSK_CONN;
   quit;
%mend orm_generateHistoryTable;


%macro orm_pivot_datatype(dtfmt=NLDATE, 
                          datatype=, 
                          in_tablenm=, 
                          out_tablenm=, 
                          debug=N);
  
  %put NOTE: {START: &sysmacroname};     
   
    /* Check the &datatype length in order to construct the &out_tablenm_live name */
    %if %length(&datatype) <= 27 %then %let datatype_len=%length(&datatype);
    %else %let datatype_len=27;

    %let out_tablenm_live=typel%substr(&datatype,1,&datatype_len);

    /* Get the &in_tablenm for the custom_objects if they have been renamed */
    /* in the detail mart */
    
    /* Create a variable (old_in_tablenm) to hold the orig. &in_tablenm value */
    /* Used by the Cust_Obj_1 -10 tables. */
    %global old_in_tablenm;
    %let old_in_tablenm=;
    
    %if %length(&in_tablenm) <= 11 and &in_tablenm ne %str() %then
    %do;
      %if %substr(&in_tablenm,1,8)=CUST_OBJ %then
      %do;
        %let k=%scan(&in_tablenm,-1,%str(_));
        %if &&cust_bus_obj_&k ne %str() %then
        %do;
          /* Put the new custom object name in the existing &in_tablenm variable */
          %let old_in_tablenm=&in_tablenm;
          %let in_tablenm=&&cust_bus_obj_&k;
        %end; 
         
        %if %sysfunc(exist(opdetail.&in_tablenm)) = 0 %then
        %do;
          %put %sysfunc(sasmsg(sashelp.ormonrpt, custom_field_no_input_tbl, NOQUOTE));
          %goto mexit;
        %end;
      %end;
    %end;
    
    %let tablenm_initial=%lowcase(%substr(&in_tablenm,1,1));

    filename &tablenm_initial.tmppgm1 temp;
    filename &tablenm_initial.tmppgm2 temp;


    /* Get all datatype custom columns in order with type information and label */
    
    %let cust_def_cnt=0;
    
    proc sql noprint;
      create table work.custom_field_defn as
      select distinct a.cust_field_nm, a.cust_field_type_cd, a.business_object_type_nm, a.cust_field_max_length,
            a.cust_field_desc
      from sasoprsk.cust_field_defn a
      where upcase(a.business_object_type_nm) = "%upcase(&datatype)" and
            a.record_version_no = (select max(b.record_version_no)
                                   from sasoprsk.cust_field_defn b
                                   where a.cust_field_nm=b.cust_field_nm and
                                         a.business_object_type_nm=b.business_object_type_nm) and
            a.valid_to_dttm=&ORM_INFINITY
      order by a.cust_field_nm;     
      
      select count(*) into : cust_def_cnt
      from work.custom_field_defn;
    quit;
    
    %if &cust_def_cnt = 0 %then
    %do;
      /* Create a special output tablenm for the business object types that belong to the same business object table */
      /* and apply additional subsetting to split up the objects into separate tables */
      /* OR use the &in_tablenm for all other objects */
      %if &out_tablenm ne %then
      %do; 
        /* Create temp_&out_tablenm tables to store the data in the database */
        %if %upcase(&out_tablenm) = TYPESCENARIO or %upcase(&out_tablenm) = TYPEASSESSMENT %then
        %do;
          proc sql;        
            create table opdetail.&out_tablenm as
            select a.*
            from opdetail.&in_tablenm a join
                 opdetail.qnr_templt b
            on a.qnr_templt_rk=b.qnr_templt_rk
              
            %if %upcase(&out_tablenm) = TYPESCENARIO %then
             %do;
               where upcase(b.scenario_flg)='Y';
             %end;
             %else %if %upcase(&out_tablenm) = TYPEASSESSMENT %then
             %do;
               where upcase(b.scenario_flg)='N' and upcase(a.template_flg)='N'
               UNION
               select *
               from opdetail.&in_tablenm 
               where upcase(template_flg)='N' and qnr_templt_rk=.;
             %end;
            quit;
        %end;
           
        %if &syserr > 4 %then
        %do;
          %put %sysfunc(sasmsg(sashelp.ormonrpt, custom_field_no_output_tbl, NOQUOTE));
          %goto mexit;
        %end;
              
        %if %upcase(&out_tablenm) ne TYPESCENARIO and %upcase(&out_tablenm) ne TYPEASSESSMENT %then
        %do; 
          /* Drop the &out_tablenm tables before appending the data to avoid duplicate record entries */
          %if %eval(%sysfunc(exist(opdetail.&out_tablenm)) = 1) %then 
          %do; 
            proc sql;
              drop table opdetail.&out_tablenm;  
            quit;
          %end; 
          proc append base=opdetail.&out_tablenm force nowarn
                                                              
          /* Hardcoded where clause used to subset the &in_tablenm when the &in_tablenm contains several */
          /* business object types */
          %if %upcase(&in_tablenm) = FINANCIAL_IMPACT %then
          %do;
            %if %upcase(&out_tablenm) = TYPERECOVERY %then
            %do;
              data=opdetail.&in_tablenm (where=(upcase(impact_type_cd)='RECOV'));
            %end;
            %else %if %upcase(&out_tablenm) = TYPEALLOCATION %then
            %do;
              data=opdetail.&in_tablenm (where=((upcase(impact_type_cd)='LOSS' or upcase(impact_type_cd)='GAIN') and upcase(is_allocation_flg)='Y'));
            %end;
            %else %if %upcase(&out_tablenm) = TYPEEFFECT %then
            %do;
              data=opdetail.&in_tablenm (where=((upcase(impact_type_cd)='LOSS' or upcase(impact_type_cd)='GAIN') and upcase(is_allocation_flg)='N'));
            %end;
          %end;
                
           %if %upcase(&in_tablenm) = ASSESSMENT %then
           %do;
             %if %upcase(&out_tablenm) = TYPEASSESSMENTTEMPLATE %then
             %do;
               data=opdetail.&in_tablenm (where=(upcase(template_flg)='Y'));
             %end;
          %end;
          run;
        %end;  
      
        %if &syserr > 4 %then
        %do;
          %put %sysfunc(sasmsg(sashelp.ormonrpt, custom_field_no_output_tbl, NOQUOTE));
          %goto mexit;
        %end;

        /* Drop the &out_tablenm_live tables before appending the data to avoid duplicate record entries */
        %if %eval(%sysfunc(exist(opdetail.&out_tablenm_live)) = 1) %then 
        %do; 
          proc sql;            
            drop table opdetail.&out_tablenm_live;  
          quit;
        %end;
           
        /* Append the &out_tablenm history table records to the live tables */
         proc append 
         %if &out_tablenm ne %str() %then
         %do;
           %let out_tblnm_len=%length(&datatype);
           
           %if &out_tblnm_len < 32 %then
           %do;
             %let out_tablenm_live=typel%substr(&datatype,1,&out_tblnm_len);
           %end;
           %else
           %do;
             %let out_tablenm_live=typel%substr(&datatype,1,32);
           %end;
           base=opdetail.&out_tablenm_live 
           data=opdetail.&out_tablenm(where=(valid_to_dttm=&future_timestamp_sas)) force nowarn;
         %end;
         %else
         %do;
           base=opdetail.&in_tablenm._l
           data=opdetail.&in_tablenm(where=(valid_to_dttm=&future_timestamp_sas)) force nowarn;
         %end;
        run;
      %end;
    %end;
    
    %if &cust_def_cnt > 0 %then
    %do;   
      /*Begin to add label to custom fields*/
      %if %sysfunc(exist(opcntl.cntl_locale))=0 %then 
      %do;
        data opcntl.cntl_locale;
         infile OPCNTLD(locale.txt) firstobs=2;
         length locale $5;
         input locale $;
        run;     
      %end;  
      
      data cntl_locale;
          set opcntl.cntl_locale;
         if locale="zh" then locale="zh_CN";
         if locale="zt" then locale="zh_TW";    
      run;

      proc sql noprint;
           select locale into :LOCALES from cntl_locale;
      quit;

      %let localecd=&LOCALES;
      
      /* Merge the default columns and custom field labels into one dataset */
      %orm_get_cust_field_labels;
      
      data work.custom_field_defn;
        length messageid $200;
        set work.custom_field_defn;          
        messageid="&datatype"||'_'||'OBJ'||'_'||trim(left(cust_field_nm));
      run;

      proc sql noprint;
        update Custom_field_defn 
        set CUST_FIELD_DESC=
           (select unicode(text,'esc') 
            from Custommessages 
            where upcase(Custommessages.key)=upcase(Custom_field_defn.messageid) and 
                  (upcase(Custommessages.locale)="%upcase(&localecd)" or
                   upcase(Custommessages.locale)="%substr(%upcase(&localecd),1,2)"));
      quit;  

     data Custom_field_defn(drop=messageid);
      set Custom_field_defn;
     run;
     
   /*End Add label to custom fields*/

     %let char_multiplier=1;
     
     /* get the MBCS multiplier */
     proc sql noprint;
       select CHAR_MULTIPLIER format=32. into :char_multiplier
       from opcntl.cntl_char_multiplier;     
     quit;
     
     /* Hardcoded table names due to business object names and primary keys not created using the */
     /* same naming scheme. This list should also include a list of tables if there is only one table */
     /* (meaning) there is no historical and live table. */
     %if %upcase(&in_tablenm) = ASSESSMENT_PERIOD %then
     %do;
       %let business_object_rk=PERIOD_RK;
     %end;
     %else %if %upcase(&in_tablenm) = ASSESSMENT_QUESTION_MASTER %then
     %do;
       %let business_object_rk=ASSESSMENT_QUESTION_RK;
     %end;
     %else %if %upcase(&in_tablenm) = QUESTION_GRP_MASTER %then
     %do;
       %let business_object_rk=ASSESSMENT_QUESTION_GROUP_RK;
     %end;
     %else %if %upcase(&in_tablenm) = CHANGE_REASON %then
     %do;
       %let business_object_rk=CHANGE_REASON_RK;
     %end;
     %else %if %upcase(&in_tablenm) = CUST_FIELD_DEFN %then
     %do;
       %let business_object_rk=CUST_FIELD_DEF_RK;
     %end;
     %else %if %upcase(&in_tablenm) = CUST_FIELD_VALUE %then
     %do;
       %let business_object_rk=CUST_FIELD_VAL_RK;
     %end;
     %else %if %upcase(&in_tablenm) = FAVORITE_AREA %then
     %do;
       %let business_object_rk=FAVORITE_RK;
     %end;
     %else %if %upcase(&in_tablenm) = GENERIC_COMMENT %then
     %do;
       %let business_object_rk=COMMENT_RK;
     %end;
     %else %if %upcase(&in_tablenm) = OP_RISK_ISSUE %then
     %do;
       %let business_object_rk=ISSUE_RK;
     %end;
     %else %if %upcase(&in_tablenm) = OBS_REQUEST %then
     %do;
       %let business_object_rk=OBS_REQ_RK;
     %end;
     %else %if %upcase(&in_tablenm) = ORAUSER %then
     %do;
       %let business_object_rk=USER_RK;
     %end;
     %else %if %upcase(&in_tablenm) = ORM_POSITION %then
     %do;
       %let business_object_rk=POSITION_RK;
     %end;
     %else %if %upcase(&in_tablenm) = ORM_ROLE %then
     %do;
       %let business_object_rk=ROLE_RK;
     %end;
     %else %if %upcase(&in_tablenm) = AUDIT_INST %then
     %do;
       %let business_object_rk=AUDIT_RK;
     %end;
     %else %if %upcase(&in_tablenm) = POLICY_INST %then
     %do;
       %let business_object_rk=POLICY_RK;
     %end;
     %else %if %upcase(&in_tablenm) = POTENTIAL_IMPACT %then
     %do;
       %let business_object_rk=IMPACT_RK;
     %end;
     %else %if %length(&old_in_tablenm) <= 11 and &old_in_tablenm ne %str() %then
     %do; 
       %if %substr(%upcase(&old_in_tablenm),1,8)=CUST_OBJ %then
       %do;
         %let business_object_rk=&old_in_tablenm._rk;
       %end; 
     %end;  
     %else /* Use the business_object name and _rk to construct the business_object_rk */
     %do;
       %let business_object_rk=&in_tablenm._rk;
     %end;
     
     /* Create the temp_&datatype table in order to modify the table structure */
     /* and delete the data in the opdetail.&in_tablenm and opdetail.&in_tablenm._l */
     /* tables in order to modify the table structures */
     /* Drop the opdetail.type1* tables */
     proc sql;
        %if &old_in_tablenm ne %str() %then
        %do;
          create table work.temp_&datatype like
              sasoprsk.&old_in_tablenm;
        %end;
        %else
        %do;   
          create table work.temp_&datatype like
            sasoprsk.&in_tablenm;
        %end;
          
        %if &out_tablenm eq %str() %then
        %do;
          delete * from opdetail.&in_tablenm;
          drop table opdetail.&in_tablenm._l;
        %end;
        %else
        %do;
          create table opdetail.&out_tablenm like
          opdetail.&in_tablenm;
            
          %if %eval(%sysfunc(exist(opdetail.&out_tablenm_live)) = 1) %then 
          %do; 
            drop table opdetail.&out_tablenm_live;  
          %end;
        %end;
     quit;
       
     /* Check to see if DBCS is set */
     /* If NO, set STR, OPM, OPS fields to a max of 4,000 */
     /* If YES, set STR, OPM, OPS fields to a max of 1,000 */
     %let DBCSVal=%sysfunc(getoption(DBCS));
     
     /* Create a copy of the original DB table before custom fields are added */
     /* This table will be merged with the custom fields later. */
     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
       proc sql;
          create table sasoprsk.temp_&datatype.O like
          work.temp_&datatype;
         
        create table work.temp_&datatype.T as
        select &business_object_rk, record_version_no, 
               valid_from_dttm, valid_to_dttm
        from work.temp_&datatype;
        
        create table work.temp_&datatype like
          work.temp_&datatype.T;
        quit;  
     %end;
        
     /* Alter the table structure of the temp and opdetail tables */  
     data _null_;
       %if &DBCSVal = DBCS %then
       %do;
         length sql_temp $%eval(500*&char_multiplier) allvars $%eval(1000*&char_multiplier);
       %end;
       %else
       %do;
         length sql_temp $%eval(2000*&char_multiplier) allvars $%eval(4000*&char_multiplier);
       %end;

      retain allvars ' ';
      set work.custom_field_defn end=lastrec;
      if cust_field_type_cd ne 'OPM' then allvars = trim(left(allvars))||'*'||left(cust_field_nm);
        
      file &tablenm_initial.tmppgm1;

      if _N_ = 1 then
         put 'proc sql noprint;';

      cust_field_type_cd = upcase(cust_field_type_cd);
      
      %if &out_tablenm eq %str() %then
      %do;
        put "alter table opdetail.&in_tablenm add ";
      %end;
      %else
      %do;
        put "alter table opdetail.&out_tablenm add ";
      %end;
        put cust_field_nm;

        if cust_field_type_cd = 'DTE' then 
        do;
          put "date format=&dtfmt.. informat=&dtfmt..";
        end;

        else if cust_field_type_cd = 'BOL' then 
        do;
          put 'character';
          put '(%eval(1';
          put '*';
          put "&char_multiplier";
          put '))';
        end;

        else if cust_field_type_cd in ('NUM' ) then 
        do;
          put 'numeric format=NLNUM19.2 informat=NLNUM19.2';
        end;

        else if cust_field_type_cd in ('STR') then 
        do;
          if cust_field_max_length >= 1 and cust_field_max_length <= 4000 then
          do;
            sql_temp="character(" || strip(put(cust_field_max_length * &char_multiplier,best.)) || ")";
            put sql_temp;
          end;
          else
          do;
            %if &DBCSVal = DBCS %then
            %do; 
              put 'character';
              put '(%eval(1000 *';
              put "&char_multiplier";
              put '))';
            %end;
            %else
            %do;
              put 'character';
              put '(%eval(4000 *';
              put "&char_multiplier";
              put '))';            
            %end;  
          end;
      end;
      else 
      do;
        /*Because we translate OPS and OPM Values expand to the max length so we avoid truncating the fields.*/
        %if &DBCSVal = DBCS %then
        %do; 
          put 'character';
          put '(%eval(1000 *';
          put "&char_multiplier";
          put '))';
        %end;
        %else
        %do;
          put 'character';
          put '(%eval(4000 *';
          put "&char_multiplier";
          put '))';            
        %end;  
      end;
        
      sql_temp = "label=" || quote(trim(cust_field_desc)) || ";";
      put sql_temp;       

      if lastrec then 
      do;
        call symput('fieldlst', allvars);
        put 'quit;';
      end;
     run;

     %include &tablenm_initial.tmppgm1;
 
     /* Put the newly modified table temp_&datatype table into the database */
     proc sql;
       create table sasoprsk.temp_&datatype like
       work.temp_&datatype;
     quit;

    /* Set the default table for the interleave macro to pull data from. */
    %if &old_in_tablenm ne %str() %then
    %do;
      %let interleaveinput=&old_in_tablenm;
    %end;
    %else
    %do;
      %let interleaveinput=&in_tablenm;
    %end; 
   
     /* Create a special output tablenm for the business object types that belong to the same business object table */
     /* and apply additional subsetting to split up the objects into separate tables */
     /* OR use the &in_tablenm for all other objects */
     %if &out_tablenm ne %then
     %do; 
       /* Create temp_&out_tablenm tables to store the data in the database */
       %if %upcase(&out_tablenm) = TYPESCENARIO or %upcase(&out_tablenm) = TYPEASSESSMENT %then
       %do;
         proc sql;        
          create table sasoprsk.typ_&datatype as
          select a.*
          from sasoprsk.&in_tablenm a join
               sasoprsk.qnr_templt b
          on a.qnr_templt_rk=b.qnr_templt_rk
        
          %if %upcase(&out_tablenm) = TYPESCENARIO %then
          %do;
            where upcase(b.scenario_flg)='Y';
          %end;
          %else %if %upcase(&out_tablenm) = TYPEASSESSMENT %then
          %do;
            where upcase(b.scenario_flg)='N' and upcase(a.template_flg)='N'
            UNION
            select *
            from sasoprsk.&in_tablenm 
            where upcase(template_flg)='N' and qnr_templt_rk=.;
          %end;
         quit;
       
       /* Set the input table name for our new interleave macro. */
      %let interleaveinput=typ_&datatype;
     %end;

       %if &syserr > 4 %then
       %do;
         %put %sysfunc(sasmsg(sashelp.ormonrpt, custom_field_no_output_tbl, NOQUOTE));
         %goto mexit;
       %end;
     
       %if %upcase(&out_tablenm) ne TYPESCENARIO and %upcase(&out_tablenm) ne TYPEASSESSMENT %then
       %do; 
            proc sql;
             create table sasoprsk.typ_&datatype like sasoprsk.&in_tablenm;
            quit;
         /* Push the data to a different table than the temp_ table. 
        Set that table as the input table name for our new interleave macro.*/
         %let interleaveinput=typ_&datatype;
         proc append base=sasoprsk.typ_&datatype force nowarn
                                                        
          /* Hardcoded where clause used to subset the &in_tablenm when the &in_tablenm contains several */
          /* business object types */
           %if %upcase(&in_tablenm) = FINANCIAL_IMPACT %then
           %do;
             %if %upcase(&out_tablenm) = TYPERECOVERY %then
             %do;
               data=sasoprsk.&in_tablenm (where=(upcase(impact_type_cd)='RECOV'));
             %end;
           %else %if %upcase(&out_tablenm) = TYPEALLOCATION %then
           %do;
             data=sasoprsk.&in_tablenm (where=((upcase(impact_type_cd)='LOSS' or upcase(impact_type_cd)='GAIN') and upcase(is_allocation_flg)='Y'));
           %end;
           %else %if %upcase(&out_tablenm) = TYPEEFFECT %then
           %do;
             data=sasoprsk.&in_tablenm (where=((upcase(impact_type_cd)='LOSS' or upcase(impact_type_cd)='GAIN') and upcase(is_allocation_flg)='N'));
           %end;
          %end;
          
        %if %upcase(&in_tablenm) = ASSESSMENT %then
         %do;
           %if %upcase(&out_tablenm) = TYPEASSESSMENTTEMPLATE %then
           %do;
             data=sasoprsk.&in_tablenm (where=(upcase(template_flg)='Y'));
            %end;
          %end;
        run;
      %end;  

      %if &syserr > 4 %then
      %do;
        %put %sysfunc(sasmsg(sashelp.ormonrpt, custom_field_no_output_tbl, NOQUOTE));
        %goto mexit;
      %end;
    %end;

   /*This quote character will allow us to wrap variable values in quotes for native sql being executed in the orm_generateHistoryTable macro.*/
   %let QUOTESTR=%str(%');

   %if &old_in_tablenm ne %str() %then
   %do;
     %let HISTORYTABLE=h_&old_in_tablenm;
   %end;
   %else
   %do;
     %let HISTORYTABLE=h_&in_tablenm;
   %end;
   
   %orm_generateHistoryTable(rk_column=&business_object_rk, 
                             in_table=&interleaveinput,
                             bo_name=&QUOTESTR.&datatype.&QUOTESTR.,
                             outputtable=&HISTORYTABLE);

   /*associate parent managed saves if the object is listed in the parent managed tables.*/
   filename wwtmppgm temp;
   data _null_;
      set opcntl.parent_managed_tables;
      file wwtmppgm;
      if trim(CHILD_TABLE) = "&in_tablenm" then
      do;
      QUOTEDPARENTTABLE =  trim("&QUOTESTR"||trim(PARENT_TABLE)||"&QUOTESTR");
      put 'proc sql;';
      put '%orm_db_connect;';
       put '&DBConnectSQL;';
       put 'execute(';
       %if %upcase(&orm_dbserver) = SQLSERVER %then
       %do;
         put "select mmm.* into  H_&HISTORYTABLE from (";
      %end;
      %else
      %do;
         put "create table H_&HISTORYTABLE as ";
       %if %upcase(&orm_dbserver) = DB2 %then
       %do;
             put " ( ";
       %end;
      %end;
      
       put "      select &HISTORYTABLE..*,";
       put '(CASE WHEN change_reason.change_reason_rk is not null THEN change_reason.change_reason_rk';
       put "              else &HISTORYTABLE..change_reason_rk END) as real_reason_rk";
       put "  from &HISTORYTABLE left outer join change_reason ";
       put '                         on change_reason.save_dttm = valid_to_dttm'; 
       put '                         and ' PARENT_RK_COLUMN ' = change_reason.business_object_rk'; 
       put "                         and change_reason.business_object_nm = " QUOTEDPARENTTABLE ;
       %if %upcase(&orm_dbserver) = SQLSERVER %then
       %do;
           put ") mmm";
       %end;
       %else %if %upcase(&orm_dbserver) = DB2 %then
       %do;
           put ") WITH NO DATA";
       %end;
      put ') by SASOPRSK_CONN;';

     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
         /* DB2 needs the data inserted still. */
         put 'execute(';
         put "   insert into H_&HISTORYTABLE select &HISTORYTABLE..*,";
         put '(CASE WHEN change_reason.change_reason_rk is not null THEN change_reason.change_reason_rk';
         put "              else &HISTORYTABLE..change_reason_rk END) as real_reason_rk";
         put "  from &HISTORYTABLE left outer join change_reason ";
         put '                         on change_reason.save_dttm = valid_to_dttm'; 
         put '                         and ' PARENT_RK_COLUMN ' = change_reason.business_object_rk'; 
         put "                         and change_reason.business_object_nm = " QUOTEDPARENTTABLE ;
         put   ') by SASOPRSK_CONN;';

         /* Rename the change reason column the DB2 way. */
         put 'execute(';
         put   "  update H_&HISTORYTABLE set change_reason_rk = real_reason_rk";
         put   ') by SASOPRSK_CONN;';
  
         put 'execute(';
         put   "  alter table H_&HISTORYTABLE drop column real_reason_rk";
         put   ') by SASOPRSK_CONN;';
  
     %end;
     %else 
     %do;
         /* Rename the change reason column the oracle or sqlserver way. */
         put 'execute(';
         put   "  alter table H_&HISTORYTABLE drop column change_reason_rk";
         put   ') by SASOPRSK_CONN;';

         put 'execute(';
         %if %upcase(&orm_dbserver) = SQLSERVER %then
          %do;
               put "sp_rename 'H_&HISTORYTABLE..real_reason_rk', 'change_reason_rk','COLUMN'";
           %end;
           %else
           %do;
             put " alter table H_&HISTORYTABLE rename column real_reason_rk to change_reason_rk";
           %end;
         put   ') by SASOPRSK_CONN;';
     %end;
     
      /*Move replace the old history table. */
       put 'execute(';
      put " drop table &HISTORYTABLE";
      put   ') by SASOPRSK_CONN;';
      put 'execute(';
     %if %upcase(&orm_dbserver) = SQLSERVER %then
     %do;
         put "select mmm.* into &HISTORYTABLE from H_&HISTORYTABLE mmm";
     %end;
     %else %if %upcase(&orm_dbserver) = DB2 %then
     %do;
         put " create table &HISTORYTABLE as (select * from H_&HISTORYTABLE  ) WITH NO DATA";   
     %end;
     %else
     %do;
        put " create table &HISTORYTABLE as select * from H_&HISTORYTABLE";
     %end;
      put   ') by SASOPRSK_CONN;';

     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
      /* We still need to insert the data for DB2 because it cannot create the table with data in the same statement. */
      put 'execute(';
      put " insert into &HISTORYTABLE select * from H_&HISTORYTABLE ";
      put   ') by SASOPRSK_CONN;';
     %end;

      put 'execute(';
      put " drop table H_&HISTORYTABLE";
      put   ') by SASOPRSK_CONN;';
      put 'disconnect from SASOPRSK_CONN;';
       put 'quit;';
      end;
   run;
   %include wwtmppgm;
   filename wwtmppgm clear;

   proc append base=sasoprsk.temp_&datatype data=sasoprsk.&HISTORYTABLE
      %if %upcase(&orm_dbserver) ne POSTGRESQL and %upcase(&orm_dbserver) ne SQLSERVER %then
      %do;
        (BULKLOAD=YES
         BL_OPTIONS='ERRORS=50'
         BL_DELETE_DATAFILE=YES) 
      %end;
      force nowarn;
   run;     
      
   /* Update the temp_&datatype.O table w/ all business object historical data */   
   %if %upcase(&orm_dbserver) = DB2 %then
   %do; 
     proc append base=sasoprsk.temp_&datatype.O data=sasoprsk.&HISTORYTABLE
        (BULKLOAD=YES
         BL_OPTIONS='ERRORS=50'
         BL_DELETE_DATAFILE=YES) 
        force nowarn;
      run; 
   %end;
    
   /*Clean up the assessment and FINANCIAL IMPACT custom tables created.*/
   /* But do not remove the Cust_Obj_1-10 tables */
   %if &interleaveinput ne &in_tablenm %then
   %do;
     %if &old_in_tablenm eq %str() %then
     %do;
       proc sql;
         drop table sasoprsk.&interleaveinput;
       quit;
     %end;
   %end;       

   /*Also we need to drop the history table created.*/
   proc sql;
      drop table sasoprsk.&HISTORYTABLE;
   quit;

    /* Define the business_object_nm */
    %if &old_in_tablenm ne %str() %then
    %do;
       %let business_object_nm=&old_in_tablenm;
    %end;
    %else
    %do;
       %let business_object_nm=&in_tablenm;
    %end;

    /* Initialize the OPM variable count and format flag */
    %let opm_var_cnt=0;
    %let opm_var_list=;
    %let opm_format_flg=N;
    
    /* Check to see if we have any OPM fields */
    proc sql noprint;
      create table work.custom_field_defn_opm as
      select *
      from work.custom_field_defn
      where strip(cust_field_type_cd)='OPM';
    quit;

    /* Keep only one record per custom OPM field */
    proc sort tagsort data=work.custom_field_defn_opm out=work.custom_field_defn_opm nodupkey;
      by cust_field_nm;
    run;
 
    proc sql noprint;
      select count(*) into : opm_var_cnt
      from work.custom_field_defn_opm;
    quit;

    %if &opm_var_cnt > 0 %then
    %do;
      proc sql;
      create table work.cust_field_value_opm as
      %if %upcase(&in_tablenm) ne CHANGE_REASON %then
       %do;
         select r.&business_object_rk, v.business_object_nm, v.row_no, 
              v.valid_from_dttm, v.valid_to_dttm, v.cust_field_nm, v.value_cd, r.record_version_no
         from sasoprsk.temp_&datatype r,
           sasoprsk.cust_field_value v, 
           work.custom_field_defn_opm d
       %end;
       %else
       %do;
         select r.&business_object_rk, v.row_no, v.cust_field_nm, d.cust_field_type_cd,
                v.value_cd
         from sasoprsk.temp_&datatype r,
           sasoprsk.cust_field_value v, 
           work.custom_field_defn_opm d
       %end;
       
       where r.&business_object_rk = v.business_object_rk
        
         %if %upcase(&in_tablenm) ne CHANGE_REASON %then
         %do;
           and v.valid_from_dttm <= r.valid_from_dttm
           and v.valid_to_dttm >= r.valid_to_dttm
           %if &old_in_tablenm ne %str() %then
           %do;
             and upcase(v.business_object_nm) = "%upcase(&old_in_tablenm)"
           %end;
           %else
           %do;
             and upcase(v.business_object_nm) = "%upcase(&in_tablenm)"
           %end;
          
         %end;
         %else
         %do;
           and upcase(v.business_object_nm) = "%upcase(&in_tablenm)"
         %end;
        
         and upcase(d.business_object_type_nm) = "%upcase(&datatype)"
         and v.cust_field_nm = d.cust_field_nm

         %if %upcase(&in_tablenm) ne CHANGE_REASON %then
         %do;
           order by r.&business_object_rk, r.record_version_no, d.cust_field_type_cd, v.cust_field_nm, v.row_no;
         %end;
         %else
         %do;
           order by r.&business_object_rk, d.cust_field_type_cd, v.cust_field_nm, v.row_no;
         %end;
       quit; 

     proc sql noprint;
       select distinct(cust_field_nm), count(*) into : opm_var_list separated by '*', 
                                                     : opmcount
       from work.cust_field_value_opm;
     quit; 

      /* If there are OPM values for a given business object */
      %if &opmcount > 0 %then
      %do;

      %let opm_format_flg=Y;
      %let i=1;
      %let h=1;
      %let opm_field_nm=%scan(&opm_var_list,&i,"*");
      %let opm_field_nms=%scan(&opm_var_list,&h,"*");
      
     /* Set OPM fields equal to value_cd */
     data work.cust_field_value_opm (drop=value_cd);
      length 
        %do %while (&opm_field_nms ne);
           &opm_field_nms
           %let h=%eval(&h+1);
           %let opm_field_nms=%scan(&opm_var_list,&h,"*");
        %end; 
         %if &DBCSVal = DBCS %then
         %do;  
           $%eval(1000*&char_multiplier);
         %end;
         %else
         %do;
           $%eval(4000*&char_multiplier);
         %end;
       
      set work.cust_field_value_opm;
      
      %do %while (&opm_field_nm ne);
        if cust_field_nm="&opm_field_nm" then &opm_field_nm=value_cd;
        %let i=%eval(&i+1);
        %let opm_field_nm=%scan(&opm_var_list,&i,"*");
      %end;          
     run; 
            
      proc sort tagsort data=cust_field_value_opm;
        by cust_field_nm;
      run;
      
      /* Apply formats to OPM fields only. We want to format the individual value cds before we concatenate them */
      %orm_set_named_list_formats(table=work.cust_field_value_opm,
                                  busobjtype=&datatype);
             
      /* Create field to hold the OPM format name so we don't lose the formatted values 
      when concatenating the multi-values into single rows */
      data _null_;
        set work.cust_field_value_opm end=lastrec;
              
        file &tablenm_initial.tmppgm2;
              
        by cust_field_nm;
        custom_field_nm="'"||strip(cust_field_nm)||"'";
        
        if _N_=1 then
        do;
         put 'data work.cust_field_value_opm;';
         put 'set work.cust_field_value_opm;';
         put 'length format_name $36;';
        end;
              
        if first.cust_field_nm then
        do;            
         
          put 'if cust_field_nm=';
          put custom_field_nm;
          put 'then format_name = ';
          put 'vformatn(';
          put cust_field_nm;
          put ');';
          
       end; 
       if lastrec then 
       do;
        put 'run;';
       end;
      run;   
      
      %include &tablenm_initial.tmppgm2;
        
      data _null_;
        set work.cust_field_value_opm end=lastrec;
        
        file &tablenm_initial.tmppgm2;
        
        by cust_field_nm;
        custom_field_nm="'"||strip(cust_field_nm)||"'";      
        format_nm = "'"||strip(format_name)||".'";
        
        if _N_ = 1 then
        do;
          put 'data work.custom_field_defn_opm;';
          put 'set work.custom_field_defn_opm;';
          put 'length format_name $36;';
        end;
        
        if first.cust_field_nm then
        do;
            
          put 'if cust_field_nm=';
          put custom_field_nm;
          put 'then format_name = ';
          put format_nm;
          put ';';
          
        end;
        
        if lastrec then
        do;
          put 'run;';
        end;
       
      run;

      %include &tablenm_initial.tmppgm2;   
                   
    /* Process the work.cust_field_value_opm table to put the formatted multi-values in a single row */
     proc sort tagsort data=work.cust_field_value_opm;
         by &business_object_rk record_version_no cust_field_nm valid_from_dttm valid_to_dttm row_no;
     run;
      
     data work.cust_field_value_opm;
      %if &DBCSVal = DBCS %then
      %do;
        length tmp_opm_value concat_values $1000;
      %end;
      %else
      %do;
        length tmp_opm_value concat_values $4000;
      %end;
       set work.cust_field_value_opm;
      %if %upcase(&in_tablenm) ne CHANGE_REASON %then
       %do;
         by &business_object_rk record_version_no cust_field_nm valid_from_dttm valid_to_dttm row_no;
       %end;
       %else
       %do;
         by &business_object_rk cust_field_nm row_no;
       %end;
       retain tmp_opm_value;
       custom_field_nm=strip(cust_field_nm);
       if first.cust_field_nm and last.cust_field_nm then 
       do;
         tmp_opm_value = strip(vvaluex(custom_field_nm)) ||",";
         concat_values = tmp_opm_value;
         output;
       end;

       else if first.cust_field_nm and not last.cust_field_nm then 
       do;
         tmp_opm_value = strip(vvaluex(custom_field_nm));
       end;

       else if not first.cust_field_nm and not last.cust_field_nm then 
      do;
        tmp_opm_value=strip(tmp_opm_value)||","|| strip(vvaluex(custom_field_nm)); 
      end;

      else if not first.cust_field_nm and last.cust_field_nm then 
      do;
        tmp_opm_value=strip(tmp_opm_value)||","|| strip(vvaluex(custom_field_nm));
        concat_values=tmp_opm_value;
        output;
      end;
     run; 
     
     /* Copy work.cust_field_value_opm_fnl to the database to perform updates */
     proc sql;
       create table sasoprsk.temp_opm_vals like work.cust_field_value_opm;
     quit;
        
     proc append base=sasoprsk.temp_opm_vals data=work.cust_field_value_opm force nowarn;
     run; 
        
     /* Update the database with the OPM field values for all parent history records */
     data _null_;
       set work.custom_field_defn_opm;
    
       file &tablenm_initial.tmppgm2;
      
       custom_field_nm="'"||strip(cust_field_nm)||"'";
       custom_field_nm_def=strip(cust_field_nm);

       put 'proc sql noprint;';
       put '%orm_db_connect;';
       put '&DBConnectSQL;';
               
       %if %upcase(&orm_dbserver) = SQLSERVER %then
       %do;
         put 'execute(';
         put "select tmp2.* into temp_&datatype.2 from (";
         put "select temp_&datatype..*, temp_opm_vals.concat_values as "; 
       %end;
       %else %if %upcase(&orm_dbserver) = DB2 %then
       %do;  
         put 'execute(';
         put "create table temp_&datatype.2 as";
         put '(';
         put "select temp_&datatype..*,temp_opm_vals.concat_values as "; 
       %end;
       %else
       %do;
         put 'execute(';
         put "create table temp_&datatype.2 as";
         put "select temp_&datatype..*, temp_opm_vals.concat_values as "; 
       %end;
        
         put cust_field_nm;
         put "from temp_&datatype left join";
         put 'temp_opm_vals';
         put 'on';
         put "temp_&datatype..&business_object_rk = temp_opm_vals.&business_object_rk and";
      
         %if %upcase(&in_tablenm) ne CHANGE_REASON %then
         %do;
           put 'temp_opm_vals.valid_from_dttm <= temp_&datatype..valid_from_dttm ';
           put 'and temp_opm_vals.valid_to_dttm > temp_&datatype..valid_from_dttm ';
           put 'and temp_opm_vals.record_version_no = temp_&datatype..record_version_no and ';
         %end;
              
         put 'temp_opm_vals.business_object_nm = ';
         put "'";
         put "&business_object_nm";
         put "'";
         put 'and cust_field_nm ='; 
         put custom_field_nm;                      
        
         %if %upcase(&orm_dbserver) = SQLSERVER %then
         %do;
           put ') tmp2';
         %end;
         %else %if %upcase(&orm_dbserver) = DB2 %then
         %do;
           put ') WITH NO DATA';
         %end;
          
         put ') by SASOPRSK_CONN;';
        
         /* DB2 creates the temp table without data so we need to populate the table still. */
         %if %upcase(&orm_dbserver) = DB2 %then
         %do;
           put 'execute (';
           put "insert into temp_&datatype.2";
           put "select temp_&datatype..*,temp_opm_vals.concat_values as ";  
           put cust_field_nm;
           put "from temp_&datatype left join";
           put 'temp_opm_vals';
           put 'on';
           put "temp_&datatype..&business_object_rk = temp_opm_vals.&business_object_rk and";
        
           %if %upcase(&in_tablenm) ne CHANGE_REASON %then
           %do;
             put 'temp_opm_vals.valid_from_dttm <= temp_&datatype..valid_from_dttm ';
             put 'and temp_opm_vals.valid_to_dttm > temp_&datatype..valid_from_dttm ';
             put 'and temp_opm_vals.record_version_no = temp_&datatype..record_version_no and ';
           %end;
              
           put 'temp_opm_vals.business_object_nm = ';
           put "'";
           put "&business_object_nm";
           put "'";
           put 'and cust_field_nm ='; 
           put custom_field_nm;   
           put ') by SASOPRSK_CONN;'; 
   
           put 'execute(';
           put "drop table temp_&datatype";    
           put ') by SASOPRSK_CONN;';
         
           put 'execute(';
           put "rename table temp_&datatype.2 to temp_&datatype";    
           put ') by SASOPRSK_CONN;';
                
         %end;
        
         %if %upcase(&orm_dbserver) ne DB2 %then
         %do;
           put 'execute(';
           put "truncate table temp_&datatype";    
           put ') by SASOPRSK_CONN;';
                     
           put 'execute(';
           put "drop table temp_&datatype";    
           put ') by SASOPRSK_CONN;';
         
           put 'execute(';
           %if %upcase(&orm_dbserver) = SQLSERVER %then
           %do;
             put "sp_rename temp_&datatype.2, temp_&datatype"; 
           %end;
           %else
           %do;
             put "alter table temp_&datatype.2 rename to temp_&datatype"; 
           %end;
           put ') by SASOPRSK_CONN;';
                        
         %end;                   
                 
        put 'disconnect from SASOPRSK_CONN;';
        put 'quit;';
      run;

        %include &tablenm_initial.tmppgm2;
                  

       /* Re-set format flag for non-OPM fields */
       %let opm_format_flg=N;
       
       proc sql;
         drop table sasoprsk.temp_opm_vals; 
       quit;

     %end; /* OPM values present */ 
   %end; /* OPM_var_cnt > 0 */


   /* Create table of custom field values for the NON-OPM fields */  
   proc sql;
     create table work.cust_field_value as
     select *
     from work.custom_field_defn
     where cust_field_type_cd ne 'OPM';
   quit;
   
   /* Keep only one record per custom field */
   proc sort tagsort data=work.cust_field_value out=work.cust_field_list nodupkey;
     by cust_field_nm;
   run;
  
   %if &out_tablenm ne %str() %then
   %do;
     %let detailtable=opdetail.&out_tablenm; 
   %end;
   %else
   %do;
     %let detailtable=opdetail.&in_tablenm;
   %end;
   
   data _null_;
     set work.cust_field_list end=lastrec;
    
     file &tablenm_initial.tmppgm2;
      
     cust_field_type_cd = upcase(cust_field_type_cd);
     custom_field_nm="'"||strip(cust_field_nm)||"'";
     custom_field_nm_def=strip(cust_field_nm);
     
     put 'proc sql noprint;';
     put '%orm_db_connect;';
     put '&DBConnectSQL;';
      
     %if %upcase(&orm_dbserver) = SQLSERVER %then
     %do;
       put 'execute(';
       put "select tmp2.* into temp_&datatype.2 from (";
       put "select temp_&datatype..*,";
     %end;
     %else %if %upcase(&orm_dbserver) = DB2 %then
     %do;
       put 'execute(';     
       put "create table temp_&datatype.2 as";
       put '(';
       put "select temp_&datatype..*,";
     %end;
     %else
     %do;
       put 'execute('; 
       put "create table temp_&datatype.2 as";
       put "select temp_&datatype..*,";
     %end;
     
     if cust_field_type_cd in ('STR') then
     do;
       put 'cust_field_value.value_txt as';
     end;
     else if cust_field_type_cd in ('OPS') then
     do;
       put 'cust_field_value.value_cd as';
     end;
     else if cust_field_type_cd in ('BOL') then
     do;
       put 'cust_field_value.value_flg as';
     end;
     else if cust_field_type_cd in ('NUM') then
     do;
       put 'cust_field_value.value_no as';
     end;
     else if cust_field_type_cd in ('DTE') then
     do;
       put 'cust_field_value.value_dt as';      
     end; 
      
     put cust_field_nm;
      
     put "from temp_&datatype left join";
     put 'cust_field_value';
     put 'on';
     put "temp_&datatype..&business_object_rk = cust_field_value.business_object_rk and";
      
     %if %upcase(&in_tablenm) ne CHANGE_REASON %then
     %do;
       put "cust_field_value.valid_from_dttm <= temp_&datatype..valid_from_dttm";
       put "and cust_field_value.valid_to_dttm > temp_&datatype..valid_from_dttm and";  
     %end;
              
     put 'cust_field_value.business_object_nm = ';
     put "'";
     put "&business_object_nm";
     put "'";
     put 'and cust_field_value.cust_field_nm ='; 
     put custom_field_nm;                      
                    
    %if %upcase(&orm_dbserver) = SQLSERVER %then
    %do;
      put ') tmp2';
    %end;
    %else %if %upcase(&orm_dbserver) = DB2 %then
    %do;
      put ') WITH NO DATA';
    %end;

    put ') by SASOPRSK_CONN;'; 
     
    /* DB2 creates the temp table without data so we need to populate the table still. */
    %if %upcase(&orm_dbserver) = DB2 %then
    %do;
      put 'execute (';
      put "insert into temp_&datatype.2";
      put "select temp_&datatype..*,";
      
      if cust_field_type_cd in ('STR') then
      do;
        put 'cust_field_value.value_txt as';
      end;
      else if cust_field_type_cd in ('OPS') then
      do;
        put 'cust_field_value.value_cd as';
      end;
      else if cust_field_type_cd in ('BOL') then
      do;
        put 'cust_field_value.value_flg as';
      end;
      else if cust_field_type_cd in ('NUM') then
      do;
        put 'cust_field_value.value_no as';
      end;
      else if cust_field_type_cd in ('DTE') then
      do;
        put 'cust_field_value.value_dt as';      
      end; 
 
      put cust_field_nm;  
      
      put "from temp_&datatype left join";
      put 'cust_field_value';
      put 'on';
      put "temp_&datatype..&business_object_rk = cust_field_value.business_object_rk and";
      
      %if %upcase(&in_tablenm) ne CHANGE_REASON %then
      %do;
        put "cust_field_value.valid_from_dttm <= temp_&datatype..valid_from_dttm";
        put "and cust_field_value.valid_to_dttm > temp_&datatype..valid_from_dttm and";  
      %end;
              
      put 'cust_field_value.business_object_nm = ';
      put "'";
      put "&business_object_nm";
      put "'";
      put 'and cust_field_value.cust_field_nm ='; 
      put custom_field_nm;  
      put ') by SASOPRSK_CONN;'; 
       
      put 'execute(';
      put "drop table temp_&datatype";    
      put ') by SASOPRSK_CONN;';
         
      put 'execute(';
      put "rename table temp_&datatype.2 to temp_&datatype";    
      put ') by SASOPRSK_CONN;';
                     
      if lastrec then
      do;
        /* Append the temp_&datatype.O and temp_&datatype tables to the opdetail library */
        /* for all history tables */
        put "create table work.temp_&datatype.F as";
        put 'select * from connection to SASOPRSK_CONN';
        put "(select temp_&datatype.O.*";
      
        %let i=1;
        %let fieldnm=%scan(&fieldlst,&i,"*");
         
        %do %while (&fieldnm ne);
          put ",temp_&datatype..&fieldnm";
          %let i=%eval(&i+1);
          %let fieldnm=%scan(&fieldlst,&i,"*");
        %end;  
        
        %let j=1;
        %let opmvarnm=%scan(&opm_var_list,&j,"*");
         
        %do %while (&opmvarnm ne);
          put ",temp_&datatype..&opmvarnm";
          %let j=%eval(&j+1);
          %let opmvarnm=%scan(&opm_var_list,&j,"*");
        %end;   
      
        put " from temp_&datatype.O left join";
        put " temp_&datatype";
        put "on temp_&datatype.O.&business_object_rk=temp_&datatype..&business_object_rk and";
        put "   temp_&datatype.O.record_version_no=temp_&datatype..record_version_no);";
      end;
    %end;
           
    %if %upcase(&orm_dbserver) ne DB2 %then
    %do;
      put 'execute(';
      put "truncate table temp_&datatype";    
      put ') by SASOPRSK_CONN;';
                     
      put 'execute(';
      put "drop table temp_&datatype";    
      put ') by SASOPRSK_CONN;';
         
      put 'execute(';
      %if %upcase(&orm_dbserver) = SQLSERVER %then
      %do;
        put "sp_rename temp_&datatype.2, temp_&datatype"; 
      %end;
      %else
      %do;
        put "alter table temp_&datatype.2 rename to temp_&datatype"; 
      %end;
      put ') by SASOPRSK_CONN;';            
    %end; 
         
      put 'disconnect from SASOPRSK_CONN;';
      put 'quit;';
    run;

   %include &tablenm_initial.tmppgm2;
    
   /* Append the temp_&datatype table to the opdetail library */
   /* for all history tables */
   proc append 
     base=&detailtable 
     
     %if %upcase(&orm_dbserver) ne DB2 %then
     %do;
        data=sasoprsk.temp_&datatype 
     %end;
     %else %if %upcase(&orm_dbserver) = DB2 and %sysfunc(exist(work.temp_&datatype.F)) %then
     %do;
        data=work.temp_&datatype.F
     %end;
        
     %if %upcase(&orm_dbserver) ne POSTGRESQL and %upcase(&orm_dbserver) ne SQLSERVER and %upcase(&orm_dbserver) ne DB2 %then
     %do;
       (BULKLOAD=YES
        BL_OPTIONS='ERRORS=50'
        BL_DELETE_DATAFILE=YES) 
     %end;
     force nowarn;
   run; 
   
  %orm_set_named_list_formats(table=&detailtable,
                              busobjtype=&datatype);
                              
  /* Correct the date and timestamp formats in the SAS datasets */
  /* for all history tables */
  %let epoch_local=0;
  proc sql noprint;
    select EPOCH_LOCAL format=32. into :epoch_local
      from opcntl.cntl_time_difference;
  quit;
     
  %orm_live_now_setup(ds=opcntl.cntl_future_timestamp);
  
  %macro orm_columns(dnames=,
                     dtnames=);
  
    %local _ndx col;
  
    %if &OPDATES eq DTTM %then %do;
      %let _ndx = 1;
      %let col = %scan(&dnames, &_ndx,"*");
      %do %while (&col ne);
        &col = datepart(&col);
        %let _ndx = %eval(&_ndx+1);
        %let col = %scan(&dnames, &_ndx,"*");
      %end;
    %end;
  
    %let _ndx = 1;
    %let col = %scan(&dtnames, &_ndx,"*");
    %do %while (&col ne);
      if &col = &future_timestamp then
        /* If the Epoch date in the macro variable future_timstamp
           is 253717660800 which is Tue, 29 Dec 10009 00:00:00 UTC
           but it shows up in the detail/report mart as 31Dec9999
           and you wonder why, read the comments below about the
           years 4000 and 8000. */
        &col = &future_timestamp_sas;
      else do;
        /* Convert Epoch date into SAS date:
           The macro variable col contains milliseconds since
           01Jan1970 UTC.  Convert into seconds.  Also, since SAS dates
           start on 01Jan1960 and are in local time, we must add a
           decade's worth of seconds plus the time difference all of
           which is stored in the macro variable epoch_local. */
        &col = (&col / 1000) + &epoch_local;
        /* The SAS date and datetime algorithm considers the
           year 4000 and the year 8000 not to be leap years.
           Subtract a day's worth of seconds for each of these
           years as necessary.  If a user enters 29Feb4000 or
           29Feb8000 in the UI then they will show up in the
           detail/report mart as 01Mar4000 and 01Mar8000.
           S0413317 */
        _tempdate_ = datepart(&col);
        if _tempdate_ > '01Mar4000'd then &col = &col - 86400;
        if _tempdate_ > '01Mar8000'd then &col = &col - 86400;
        drop _tempdate_;
      end;
      %let _ndx = %eval(&_ndx+1);
      %let col = %scan(&dtnames, &_ndx,"*");
    %end;
  
  %mend orm_columns;
  
  %let dtvars=;
  %let dtvars1=;
  %let dtvars2=;
  %let dttmvars=;
  
  /* Put the list of date and timestamp columns into macro variables */
  proc sql noprint;
    select column_name into : dtvars1 separated by "*"
      from opcntl.cntl_dm_columns
      where column_type='DATE' and 
      %if &old_in_tablenm ne %str() %then
      %do;
        upcase(base_table)="%upcase(&old_in_tablenm)" 
      %end;
      %else
      %do;
        upcase(base_table)="%upcase(&business_object_nm)" 
      %end;  
      ;
     
    select cust_field_nm into : dtvars2 separated by "*"
      from work.custom_field_defn
      where upcase(cust_field_type_cd)='DTE';
    select column_name into : dttmvars separated by "*"
      from opcntl.cntl_dm_columns
     where column_type='TIMESTAMP' and 
      %if &old_in_tablenm ne %str() %then
      %do;
        upcase(base_table)="%upcase(&old_in_tablenm)" 
    %end;
    %else
    %do;
       upcase(base_table)="%upcase(&business_object_nm)" 
    %end;
      ;
    
  quit;
   
  %if %nrbquote(&dtvars2) ne  %then
  %do;
    %let dtvars=&dtvars1*&dtvars2;
  %end;
  %else
  %do;
    %let dtvars=&dtvars1;
  %end;
   
  /* Read through the columns reformat the date and timestamp columns */
  %if &out_tablenm ne %str() %then
  %do;
    data opdetail.&out_tablenm;
      set opdetail.&out_tablenm; 
      %orm_columns(dnames=&dtvars,
                   dtnames=&dttmvars);
    run;
  %end;
  %else
  %do;
    data opdetail.&in_tablenm;
      set opdetail.&in_tablenm; 
      %orm_columns(dnames=&dtvars,
                   dtnames=&dttmvars);
    run;
  %end; 
  
   /* Append the temp_&datatype table to the opdetail library */
   /* for the live tables */
   proc append 
     %if &out_tablenm ne %str() %then
     %do;
       base=opdetail.&out_tablenm_live 
       data=opdetail.&out_tablenm(where=(valid_to_dttm=&future_timestamp_sas)) force nowarn;
     %end;
     %else
     %do;
       base=opdetail.&in_tablenm._l
       data=opdetail.&in_tablenm(where=(valid_to_dttm=&future_timestamp_sas)) force nowarn;
     %end;
   run;
%end;
  
%mexit:

   %if %upcase(&debug)=N %then
   %do;
     filename &tablenm_initial.tmppgm1 clear;
     filename &tablenm_initial.tmppgm2 clear;

    
     %if %eval(%sysfunc(exist(sasoprsk.temp_&datatype)) = 1) %then 
     %do;
       proc sql;
         drop table sasoprsk.temp_&datatype; 
       quit;
     %end;
    
     %if %upcase(&orm_dbserver) = DB2 %then
     %do;
       %if %eval(%sysfunc(exist(sasoprsk.temp_&datatype.O)) = 1) %then
       %do;
         proc sql;
           drop table sasoprsk.temp_&datatype.O;
         quit;
       %end;
     %end;
    
     proc datasets library=work nolist nowarn;
          delete custom_field_defn
                 cust_field_value/mt=all;
     run;
   %end;

%put NOTE: {END: &sysmacroname};

%mend orm_pivot_datatype;
